###
# Copyright (c) 2005, Daniel DiPaolo
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
###

from supybot.test import *

class QuoteTestCase(ChannelPluginTestCase):
    plugins = ('Quote', 'User')

    def testReplace(self):
        self.feedMsg('register testuser moo', to=self.nick, frm=self.prefix)
        _ = self.getMsg(' ')
        self.assertNotError("quote add hello")
        self.assertNotError("quote replace 1 goodbye")
        self.assertRegexp("quote get 1", "goodbye")
        self.assertError("quote replace 5 afsdafas") # non-existant

    def testUnauthenticatedAdd(self):
        # This should fail because the user isn't registered
        self.assertError('quote add hello world!')
        with conf.supybot.databases.plugins.requireRegistration.context(False):
            self.assertNotError('quote add hello world!')
            self.assertRegexp('quote get 1', 'hello')
            self.assertNotError('quote remove 1')

# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:
